import jmespath.parser
import jmespath.visitor
import jmespath.functions
import jmespath.exceptions
import itertools
import types

# Register generator type in jmespath
jmespath.functions.TYPES_MAP['generator'] = 'array'
jmespath.functions.REVERSE_TYPES_MAP['array'] += ('generator',)

# Register biopython types in jmespath
jmespath.functions.TYPES_MAP['Seq'] = 'string'
jmespath.functions.REVERSE_TYPES_MAP['string'] += ('Seq',)
jmespath.functions.REVERSE_TYPES_MAP['Seq'] = ('Seq',)
jmespath.functions.REVERSE_TYPES_MAP['SeqFeature'] = ('SeqFeature',)
jmespath.functions.TYPES_MAP['ExactPosition'] = 'number'
jmespath.functions.TYPES_MAP['BeforePosition'] = 'number'
jmespath.functions.TYPES_MAP['BetweenPosition'] = 'number'
jmespath.functions.TYPES_MAP['AfterPosition'] = 'number'
jmespath.functions.TYPES_MAP['OneOfPosition'] = 'number'
jmespath.functions.REVERSE_TYPES_MAP['number'] += ('ExactPosition',)
jmespath.functions.REVERSE_TYPES_MAP['number'] += ('BeforePosition',)
jmespath.functions.REVERSE_TYPES_MAP['number'] += ('BetweenPosition',)
jmespath.functions.REVERSE_TYPES_MAP['number'] += ('AfterPosition',)
jmespath.functions.REVERSE_TYPES_MAP['number'] += ('OneOfPosition',)

# this implementation includes https://github.com/jmespath/jmespath.site/pull/6
# and https://github.com/jmespath/jmespath.py/issues/159


def compile(expression):
    return Parser().parse(expression)


def search(expression, data, options=None):
    return Parser().parse(expression).search(data, options=options)


class Parser(jmespath.parser.Parser):
    def _parse(self, expression):
        result = super()._parse(expression)
        return ParsedResult(result.expression, result.parsed)


class ParsedResult(jmespath.parser.ParsedResult):
    def search(self, value, options=None):
        interpreter = TreeInterpreterGenerator(options)
        result = interpreter.visit(self.parsed, value)
        return result


class ExtendedFunctions(jmespath.functions.Functions):
    def call_function(self, function_name, resolved_args, **kwargs):
        try:
            spec = self.FUNCTION_TABLE[function_name]
        except KeyError:
            raise jmespath.exceptions.UnknownFunctionError(
                "Unknown function: %s()" % function_name)
        function = spec['function']
        signature = spec['signature']
        self._validate_arguments(resolved_args, signature, function_name)
        return function(self, *resolved_args, **kwargs)

    @jmespath.functions.signature({'types': ['object']}, {'types': ['expref']})
    def _func_let(self, lexical_scope, expref, **kwargs):
        if 'scope' in kwargs:
            scope = dict(kwargs['scope'])
            scope.update(lexical_scope)
        else:
            scope = dict(lexical_scope)
        kwargs['scope'] = scope
        return expref.visit(expref.expression, expref.context, **kwargs)

    @jmespath.functions.signature({'types': ['string']}, {'types': ['string']})
    def _func_split(self, on, val):
        return val.split(on)

    @jmespath.functions.signature({'types': ['Seq']}, {'types': ['SeqFeature']})
    def _func_extract(self, seq, feature):
        return feature.extract(seq)


class _Expression(jmespath.visitor._Expression):
    def __init__(self, expression, interpreter, context):
        super().__init__(expression, interpreter)
        self.context = context


class TreeInterpreterGenerator(jmespath.visitor.TreeInterpreter):
    def __init__(self, options=None, *args, **kwargs):
        options = options or jmespath.visitor.Options(custom_functions=ExtendedFunctions())
        super().__init__(*args, options=options, **kwargs)
        self._generators = {}

    def _gen_to_list(self, gen, recurse=False):
        """
        Called when a jmespath operation requires random access to generator.
        Memoises already converted generators and returns the already generated list.
        :param gen: generator object
        :return: list of values returned by generator
        """
        if getattr(gen, '__hash__', None) is not None:
            gen = self._generators.get(gen, gen)
        if isinstance(gen, types.GeneratorType):
            l = list(gen)
            self._generators[gen] = l
            if recurse:
                for i in range(len(l)):
                    if isinstance(l[i], types.GeneratorType):
                        l[i] = self._gen_to_list(l[i])
            return l
        return gen

    def visit(self, node, *args, **kwargs):
        # if a visit caused list conversion, get list. Assume that 'value' is args[0].
        if len(args) and isinstance(args[0], types.GeneratorType):
            args = list(args)  # convert from tuple
            args[0] = self._generators.get(args[0], args[0])
        return super().visit(node, *args, **kwargs)

    def visit_field(self, node, value, **kwargs):
        try:
            return value.get(node['value'])
        except AttributeError:
            # Allow accessing objects fields TODO push this change upstream, possibly with config flag
            try:
                return getattr(value, node['value'])
            except AttributeError:
                # If the field is not defined in the current object, then fall back
                # to checking in the scope chain, if there's any that has been
                # created.
                if 'scope' in kwargs:
                    return kwargs['scope'].get(node['value'], None)
                return None

    def visit_function_expression(self, node, value, **kwargs):
        resolved_args = []
        for child in node['children']:
            current = self._gen_to_list(self.visit(child, value, **kwargs), True)
            resolved_args.append(current)
        return self._functions.call_function(node['value'], resolved_args)

    def visit_not_expression(self, node, value, **kwargs):
        original_result = self.visit(node['children'][0], value, **kwargs)
        if original_result == 0:
            # Special case for 0, !0 should be false, not true.
            # 0 is not a special cased integer in jmespath.
            return False
        return self._is_false(original_result) # TODO bugfix, push this change upstream

    def visit_filter_projection(self, node, value, **kwargs):
        base = self.visit(node['children'][0], value, **kwargs)
        if not isinstance(base, (list, types.GeneratorType, map, filter)):
            return None
        comparator_node = node['children'][2]
        for element in base:
            comparison = self.visit(comparator_node, element, **kwargs)
            if self._is_true(comparison):
                current = self.visit(node['children'][1], element, **kwargs)
                if current is not None:
                    yield current

    def visit_flatten(self, node, value, **kwargs):
        base = self.visit(node['children'][0], value, **kwargs)
        if not isinstance(base, (list, types.GeneratorType, map, filter)):
            # Can't flatten the object if it's not a list.
            return None
        for element in base:
            if isinstance(element, (list, types.GeneratorType, map, filter)):
                for subelement in element:
                    yield subelement
            else:
                yield element

    def visit_index(self, node, value, **kwargs):
        value = self._gen_to_list(value)
        return super().visit_index(node, value)

    def visit_slice(self, node, value, **kwargs):
        return itertools.islice(value, *node['children'])

    def visit_multi_select_list(self, node, value, **kwargs):
        if value is None:
            return None
        for child in node['children']:
            yield self.visit(child, value, **kwargs)

    def visit_projection(self, node, value, **kwargs):
        base = self.visit(node['children'][0], value, **kwargs)
        if not isinstance(base, (list, types.GeneratorType, map, filter)):
            return None
        for element in base:
            current = self.visit(node['children'][1], element, **kwargs)
            if current is not None:
                yield current

    def visit_value_projection(self, node, value, **kwargs):
        base = self.visit(node['children'][0], value, **kwargs)
        try:
            base = base.values()
        except AttributeError:
            return None
        for element in base:
            current = self.visit(node['children'][1], element, **kwargs)
            if current is not None:
                yield current

    def _is_false(self, value, **kwargs):
        if isinstance(value, types.GeneratorType):
            # peek generator instead of _gen_to_list()
            try:
                peek = next(value)
                old_value = value
                value = itertools.chain((peek,), value)
                self._generators[id(old_value)] = value
                return False
            except StopIteration:
                return True
        return super()._is_false(value)

    def visit_expref(self, node, value, **kwargs):
        return _Expression(node['children'][0], self, value)

    def visit_subexpression(self, node, value, **kwargs):
        result = value
        for node in node['children']:
            result = self.visit(node, result, **kwargs)
        return result

    def visit_comparator(self, node, value, **kwargs):
        # Common case: comparator is == or !=
        comparator_func = self.COMPARATOR_FUNC[node['value']]
        if node['value'] in self._EQUALITY_OPS:
            return comparator_func(
                self.visit(node['children'][0], value, **kwargs),
                self.visit(node['children'][1], value, **kwargs)
            )
        else:
            # Ordering operators are only valid for numbers.
            # Evaluating any other type with a comparison operator
            # will yield a None value.
            left = self.visit(node['children'][0], value, **kwargs)
            right = self.visit(node['children'][1], value, **kwargs)
            num_types = (int, float)
            if not (jmespath._is_comparable(left) and
                    jmespath._is_comparable(right)):
                return None
            return comparator_func(left, right)

    def visit_current(self, node, value, **kwargs):
        return super().visit_current(node, value)

    def visit_identity(self, node, value, **kwargs):
        return super().visit_identity(node, value)

    def visit_index_expression(self, node, value, **kwargs):
        result = value
        for node in node['children']:
            result = self.visit(node, result, **kwargs)
        return result

    def visit_key_val_pair(self, node, value, **kwargs):
        return self.visit(node['children'][0], value, **kwargs)

    def visit_literal(self, node, value, **kwargs):
        return super().visit_literal(node, value)

    def visit_multi_select_dict(self, node, value, **kwargs):
        if value is None:
            return None
        collected = self._dict_cls()
        for child in node['children']:
            collected[child['value']] = self.visit(child, value, **kwargs)
        return collected

    def visit_or_expression(self, node, value, **kwargs):
        matched = self.visit(node['children'][0], value, **kwargs)
        if self._is_false(matched):
            matched = self.visit(node['children'][1], value, **kwargs)
        return matched

    def visit_and_expression(self, node, value, **kwargs):
        matched = self.visit(node['children'][0], value, **kwargs)
        if self._is_false(matched):
            return matched
        return self.visit(node['children'][1], value, **kwargs)

    def visit_pipe(self, node, value, **kwargs):
        result = value
        for node in node['children']:
            result = self.visit(node, result, **kwargs)
        return result

    def _is_true(self, value, **kwargs):
        return super()._is_true(value)




"""
Biopython Convert
Convert between any formats that Biopython supports or gffutils.
Provides a means of querying/filtering documents using JMESPath query language.
"""
import sys
import pathlib
import itertools
import types
from collections import defaultdict

import getopt
from typing import Callable, Generator, OrderedDict

from Bio import SeqIO, StreamModeError, SeqFeature, Seq
import gffutils
from gffutils import biopython_integration

from . import JMESPathGen

gff_types = ['gff', 'gff3']
extended_types = ['text', 'json', 'yaml', 'yml']
SeqIO_types = ['abi', 'abi-trim', 'ace', 'cif-atom', 'cif-seqres', 'clustal', 'embl', 'fasta', 'fasta-2line',
               'fastq-sanger', 'fastq', 'fastq-solexa', 'fastq-illumina', 'genbank', 'gb', 'ig', 'imgt', 'nexus',
               'pdb-seqres', 'pdb-atom', 'phd', 'phylip', 'pir', 'seqxml', 'sff', 'sff-trim', 'stockholm', 'swiss',
               'tab', 'qual', 'uniprot-xml']
stat_annotations = ['molecule_type', 'topology', 'data_file_division', 'date', 'accessions', 'sequence_version', 'gi',
                    'keywords', 'source', 'organism']

usage = """\
Use: biopython.convert [-s] [-v] [-i] [-q JMESPath] input_file input_type output_file output_type
\t-s Split records into seperate files
\t-q JMESPath to select records. Must return list of SeqIO records. Root is list of input SeqIO records.
\t-i Print out details of records during conversion
\t-v Print version and exit
""" + "\nValid types: " + ', '.join(SeqIO_types + gff_types + extended_types) + "\n"


def get_args(sysargs: list):
    """
    Parse command line arguments
    :param sysargs: list of command line arguments (sys.argv[1:])
    :return: (input_path, input_type, output_path, output_type, split, jmespath, stats)
    """
    split = False
    jpath = None
    stats = None
    # Parse arguments
    try:
        opts, args = getopt.gnu_getopt(sysargs, 'vsiq:')
        for opt, val in opts:
            if opt == '-v':
                from . import __version
                print(__version.__version__)
                exit(0)
            elif opt == '-s':
                split = True
            elif opt == '-q':
                if not val:
                    raise getopt.GetoptError("JMESPath must not be empty", "-q")
                jpath = val
            elif opt == '-i':
                stats = sys.stdout

    except getopt.GetoptError as err:
        print("Argument error(" + str(err.opt) + "): " + err.msg, file=sys.stderr)
        args = []

    # Check for minimum number of arguments
    if len(args) < 4:
        print(usage, file=sys.stderr)
        exit(1)

    input_path = pathlib.Path(args[0])
    input_type = args[1]
    output_path = pathlib.Path(args[2])
    output_type = args[3]

    return input_path, input_type, output_path, output_type, split, jpath, stats


def to_stats(record: SeqIO.SeqRecord) -> str:
    """
    Build GFF record representing summary of SeqRecord
    :param record: SeqIO.SeqRecord to represent
    :return: string containing GFF record
    """
    # 'if' statements can be removed after https://github.com/daler/gffutils/pull/144
    if record.name:
        attributes = {'Name': [record.name]}
    else:
        attributes = {}
    for k, v in record.annotations.items():
        if k in stat_annotations:
            if isinstance(v, list):
                v = [str(a) for a in v]
            else:
                v = [str(v)]
            if v:
                attributes[k] = v

    feat_count = defaultdict(int)
    for f in record.features:
        if f.type == 'source':
            # Include source coordinate
            attributes[f"source__location"] = [str(part) for part in f.location.parts]
            # Include source qualifiers
            for k, v in f.qualifiers.items():
                attr = attributes.get(f"source_{k}", [])
                if not isinstance(attr, list):
                    attr = [attr]
                attr.append(v)
                attributes[f"source_{k}"] = attr
        # Count features of each type
        feat_count[f.type] += 1
    attributes['features'] = [f"{k}:{v}" for k, v in feat_count.items()]

    if record.description:
        attributes['desc'] = [record.description]
    return str(gffutils.Feature(record.id, "biopython.convert", "sequence", start=1, end=len(record), attributes=attributes))


def _allow_single(records):
    """
    Helper to allow returing single record from JMESPath
    :param records: SeqIO.SeqRecord instance
    :return: tuple containing SeqIO.SeqRecord
    """
    if isinstance(records, SeqIO.SeqRecord):
        # Support returning single record from JMESPath
        return (records,)
    return records


def _to_SeqRecord(record):
    """
    Support generating a single new record in JMESPath
    :param record: dict of SeqRecord constructor parameters
    :return: SeqRecord
    """
    seq = record.get('seq', {})
    del record['seq']
    if isinstance(seq, str):
        seq = Seq.Seq(seq)
    elif isinstance(seq, dict):
        seq = Seq.Seq(**seq)
    return SeqIO.SeqRecord(seq=seq, **record)


def _to_SeqRecords(records):
    """
    Helper to convert all output records to SeqRecords
    :param records: dict or SeqRecord
    :return: SeqRecord
    """
    if isinstance(records, dict):
        records = _to_SeqRecord(records)

    records = _allow_single(records)

    return map(lambda r: _to_SeqRecord(r) if isinstance(r, dict) else r, records)


def get_records(input_handle, input_type: str, jpath: str = '', xform: Callable = _to_SeqRecords):
    """
    Read in records and apply optional jmespath
    :param input_handle: File handle to read data from
    :param input_type: one of abi,abi-trim,ace,cif-atom,cif-seqres,clustal,embl,fasta,fasta-2line,fastq-sanger,fastq,
        fastq-solexa,fastq-illumina,genbank,gb,ig,imgt,nexus,pdb-seqres,pdb-atom,phd,phylip,pir,seqxml,sff,sff-trim,
        stockholm,swiss,tab,qual,uniprot-xml,gff3
    :param jpath: JMESPath selecting records to keep. The root is the list of records. The path must return a list of records.
    :param xform: Callable that takes the result of the jmespath and does anything necessary to convert to a iterable of output records
    :return: iterable of resulting records
    """
    def gentype(x):
        # Newer versions of biopython return iterator that is not of type Generator, wrap in generator type so jmespath can detect
        for a in x:
            yield a

    if input_type in gff_types:
        # If input is GFF load with gffutils library
        db = gffutils.create_db(input_handle, ":memory:", merge_strategy="create_unique")
        # Wrap features in generator that converts to BioPython SeqRecords
        input_records = map(
            lambda x: SeqIO.SeqRecord("", features=list(x)),
            itertools.groupby(
                map(biopython_integration.to_seqfeature, db.all_features(order_by="seqid")),
                lambda x: x.id
            )
        )
    else:
        input_records = SeqIO.parse(input_handle, input_type)

    # Wrap input in JMESPath selector if provided
    if jpath:
        input_records = JMESPathGen.search(jpath, gentype(input_records))

    # Apply xform to both entire return value
    input_records = xform(input_records)

    return input_records


def _generate_suffixes(path: pathlib.Path) -> Generator[pathlib.Path, None, None]:
    """
    Helper to generate a new file path from a base path on each iteration
    :param path: base path to add suffix to
    :return: new path
    """
    i = 0
    while True:
        # Open output file with file name suffix if splitting
        yield path.with_suffix(f".{i}{path.suffix}")
        i += 1


def gff_writer(records: [SeqIO.SeqRecord], handle, output_type: str):
    """
    Convert SeqRecord to gffutils GFF3 record and output to handle
    :param handle: file handle to write to
    :param records: iterable of SeqRecord instances
    :param output_type: output format, ignored
    :return: None
    """
    for record in records:
        # TODO extend gffutils SeqFeature support
        for feature in record.features:
            feature = biopython_integration.from_seqfeature(feature)
            feature.seqid = record.id
            feature.source = 'biopython.convert'
            print(feature, file=handle)


def _print_stats(record, stats):
    """
    Helper to print stats of record
    :param record: SeqRecord to print stats of
    :param stats: IO handle or None
    :return: record, unaltered
    """
    if stats and isinstance(record, SeqIO.SeqRecord):
        print(to_stats(record), file=stats)
    return record


def to_strings(v):
    """
    Helper to recursively convert Generators to lists, stringifing all else
    :param v: Parent object/list
    :return: list/dict with all children converted to the same or a string
    """
    if isinstance(v, str):
        return v

    if isinstance(v, (types.GeneratorType, map, filter, tuple)):
        v = list(v)

    if hasattr(v, 'keys'):
        keys = v.keys()
    elif hasattr(v, '__getitem__'):
        keys = range(len(v))
    else:
        return str(v)

    for i in keys:
        v[i] = to_strings(v[i])
    return v


def to_dicts(v):
    """
    Helper to recursively convert Objects and Generators to dicts and lists
    :param v: Parent object/list
    :return: list/dict with all children converted to the same
    """
    if isinstance(v, str):
        try:
            return int(v)
        except ValueError:
            pass
        return v

    if isinstance(v, Seq.Seq):
        return str(v)

    if isinstance(v, (types.GeneratorType, map, filter, tuple)):
        v = list(v)

    if isinstance(v, SeqIO.SeqRecord):
        v = {
            **v.__dict__,
            'seq': str(v.seq)
        }
        del v['_seq']
    elif isinstance(v, SeqFeature.FeatureLocation):
        v = {
            **v.__dict__,
            'start': v.start,
            'end': v.end,
            'strand': v.strand,
        }
        del v['_start']
        del v['_end']
        del v['_strand']
    elif isinstance(v, SeqFeature.AbstractPosition):
        return to_dicts(str(v))
    elif isinstance(v, OrderedDict):
        v = dict(v)
    elif hasattr(v, '__dict__'):
        v = v.__dict__

    if hasattr(v, 'keys'):
        keys = v.keys()
    elif hasattr(v, '__getitem__'):
        keys = range(len(v))
    else:
        return v

    for i in keys:
        v[i] = to_dicts(v[i])
    return v


def convert(input_path: pathlib.Path, input_type: str, output_path: pathlib.Path, output_type: str, split: bool = False, jpath: str = '', stats=None):
    """
    Convert document from one format to another, optionally querying via JMESPath or splitting into separate outputs
    :param input_path: Path to input dataset
    :param input_type: Format of input dataset
    :param output_path: Path to output dataset
    :param output_type: Format of output dataset
    :param split: Split each record into a different output dataset. Adds index suffix to output path.
    :param jpath: JMESPath query to apply to input dataset before outputting
    :param stats: File handle to output GFF3 summary of output records
    :return: None
    """
    xform = _to_SeqRecords
    with input_path.open("r") as handle:
        if output_type == 'text':
            writer = lambda records, fh, t: fh.write("\n".join(map(str, to_strings(records))) + "\n")
            xform = lambda x: x
        elif output_type == 'json':
            import json
            writer = lambda records, fh, t: json.dump(to_dicts(records), fh, skipkeys=True, indent=True)
            xform = _allow_single
        elif output_type in ('yml', 'yaml'):
            from ruamel.yaml import YAML
            yml = YAML(typ='unsafe')
            writer = lambda records, fh, t: yml.dump(to_dicts(records), fh)
            xform = _allow_single
        elif output_type in gff_types:
            writer = gff_writer
        else:
            writer = SeqIO.write

        if stats:
            print("##gff-version 3", file=stats)

        seq_records = get_records(handle, input_type, jpath, xform)
        binary = ''
        if split:
            for record, path in zip(seq_records, _generate_suffixes(output_path)):
                _print_stats(record, stats)
                while True:
                    try:
                        with path.open('w' + binary) as output_handle:
                            writer((record,), output_handle, output_type)
                    except StreamModeError:
                        if binary == 'b':
                            raise
                        binary = 'b'
                        continue
                    break
        else:
            while True:
                try:
                    with output_path.open('w' + binary) as output_handle:
                        writer(
                            map(
                                lambda r: _print_stats(r, stats),
                                seq_records
                            ),
                            output_handle,
                            output_type
                        )
                except StreamModeError:
                    if binary == 'b':
                        raise
                    binary = 'b'
                    continue
                break




import difflib
import io
from unittest import TestCase
from hashlib import sha256
from tempfile import TemporaryDirectory
from pathlib import Path

from biopython_convert import convert


class TestConvert(TestCase):
    input_path = Path('test-data/has_plasmids.gbff')
    output_path = Path('test-data/outputs/')
    input_type = 'genbank'
    convert_type = 'embl'
    filter = "[?!(features[?type=='source'].qualifiers.plasmid)]"

    def setUp(self) -> None:
        self.workdir = TemporaryDirectory()

    def hash_file(self, path: Path):
        with path.open('rb') as output_handle:
            return sha256(output_handle.read()).hexdigest()

    def compare_files(self, a: Path, b: Path):
        with open(a) as a_stream:
            with open(b) as b_stream:
                diffs = list(difflib.unified_diff(a_stream.readlines(), b_stream.readlines(), fromfile=str(a), tofile=str(b)))
                self.assertEqual(0, len(diffs), ''.join(diffs))

    def tearDown(self) -> None:
        self.workdir.cleanup()

    def test_basic(self):
        output_path = Path(self.workdir.name, 'basic')
        convert(self.input_path, self.input_type, output_path, self.input_type)
        self.compare_files(Path.joinpath(self.output_path, 'basic'), output_path)

    def test_info(self):
        output_path = Path(self.workdir.name, 'basic')
        stats = io.StringIO()
        convert(self.input_path, self.input_type, output_path, self.input_type, stats=stats)
        self.compare_files(Path.joinpath(self.output_path, 'basic'), output_path)
        a = Path.joinpath(self.output_path, 'info')
        with open(a) as a_stream:
            diffs = list(difflib.unified_diff(a_stream.readlines(), stats.getvalue().splitlines(keepends=True), fromfile=str(a), tofile='stdout'))
            self.assertEqual(0, len(diffs), ''.join(diffs))

    def test_convert(self):
        output_path = Path(self.workdir.name, 'convert')
        convert(self.input_path, self.input_type, output_path, self.convert_type)
        self.compare_files(Path.joinpath(self.output_path, 'convert'), output_path)

    def test_filter(self):
        output_path = Path(self.workdir.name, 'filter')
        convert(self.input_path, self.input_type, output_path, self.input_type, jpath=self.filter)
        self.compare_files(Path.joinpath(self.output_path, 'filter'), output_path)

    def test_split(self):
        output_path = Path(self.workdir.name, 'record')
        convert(self.input_path, self.input_type, output_path, self.input_type, split=True)
        files = list(x for x in Path(self.workdir.name).glob('*') if x.is_file())
        truth_files = list(Path.joinpath(self.output_path, 'split').glob('*'))
        files.sort()
        truth_files.sort()
        self.assertListEqual([f.name for f in truth_files], [f.name for f in files])
        for a, b in zip(truth_files, files):
            self.compare_files(a, b)

    def test_gff(self):
        output_path = Path(self.workdir.name, 'gff')
        convert(self.input_path, self.input_type, output_path, 'gff3')
        self.compare_files(Path.joinpath(self.output_path, 'gff'), output_path)

    def test_txt(self):
        output_path = Path(self.workdir.name, 'txt')
        convert(self.input_path, self.input_type, output_path, 'text', jpath='[*].annotations.taxonomy')
        self.compare_files(Path.joinpath(self.output_path, 'txt'), output_path)

    def test_ptt(self):
        output_path = Path(self.workdir.name, 'ptt')
        convert(self.input_path, self.input_type, output_path, 'text', jpath="[0].[join(' - 1..', [description, to_string(length(seq))]), join(' ', [to_string(length(features[?type=='CDS' && qualifiers.translation])), 'proteins']), join(`\"\\t\"`, ['Location', 'Strand', 'Length', 'PID', 'Gene', 'Synonym', 'Code', 'COG', 'Product']), (features[?type=='CDS' && qualifiers.translation].[join('..', [to_string(sum([location.start, `1`])), to_string(location.end)]), [location.strand][?@==`1`] && '+' || '-', length(qualifiers.translation[0]), (qualifiers.db_xref[?starts_with(@, 'GI')].split(':', @)[1])[0] || '-', qualifiers.gene[0] || '-', qualifiers.locus_tag[0] || '-', '-', '-', qualifiers.product[0] ] | [*].join(`\"\\t\"`, [*].to_string(@)) )] | []")
        self.compare_files(Path.joinpath(self.output_path, 'ptt'), output_path)

    def test_txt_stats(self):
        output_path = Path(self.workdir.name, 'txt')
        stats = io.StringIO()
        convert(self.input_path, self.input_type, output_path, 'text', jpath='[*].annotations.taxonomy', stats=stats)
        self.compare_files(Path.joinpath(self.output_path, 'txt'), output_path)
        self.assertEqual(1, len(stats.getvalue().splitlines(keepends=True)))

    def test_json(self):
        output_path = Path(self.workdir.name, 'json')
        convert(self.input_path, self.input_type, output_path, 'json')
        self.compare_files(Path.joinpath(self.output_path, 'json'), output_path)

    def test_yaml(self):
        output_path = Path(self.workdir.name, 'yaml')
        convert(self.input_path, self.input_type, output_path, 'yaml')
        self.compare_files(Path.joinpath(self.output_path, 'yaml'), output_path)

    def test_json_jpath(self):
        output_path = Path(self.workdir.name, 'json_jpath')
        convert(self.input_path, self.input_type, output_path, 'json', jpath='[*].{id: id, type: annotations.molecule_type}')
        self.compare_files(Path.joinpath(self.output_path, 'json_jpath'), output_path)

    def test_json_jpath_split(self):
        output_path = Path(self.workdir.name, 'json_jpath_split')
        convert(self.input_path, self.input_type, output_path, 'json', jpath='[*].{id: split(\'.\', id), type: annotations.molecule_type}')
        self.compare_files(Path.joinpath(self.output_path, 'json_jpath_split'), output_path)

    def test_json_jpath_extract(self):
        output_path = Path(self.workdir.name, 'json_jpath_extract')
        convert(self.input_path, self.input_type, output_path, 'json', jpath="[0].let({seq:seq}, &features[?type=='gene'].extract(seq, @))") # "[0].let({seq: seq}, &features[?type=='gene']|[0].{id: id, description: 'extracted sequence', seq: seq[::location]})")
        self.compare_files(Path.joinpath(self.output_path, 'json_jpath_extract'), output_path)

    def test_json_jpath_let(self):
        output_path = Path(self.workdir.name, 'json_jpath_let')
        convert(self.input_path, self.input_type, output_path, 'json', jpath='[*].let({type: annotations.molecule_type}, &{id: id, type: type})')
        self.compare_files(Path.joinpath(self.output_path, 'json_jpath_let'), output_path)

    def test_yaml_jpath(self):
        output_path = Path(self.workdir.name, 'yaml_jpath')
        convert(self.input_path, self.input_type, output_path, 'yaml', jpath='[*].{id: id, annotations: annotations}')
        self.compare_files(Path.joinpath(self.output_path, 'yaml_jpath'), output_path)

    def test_gentype(self):
        """
        Test handling the generator type within JMESPath functions
        """
        output_path = Path(self.workdir.name, 'gentype')
        convert(self.input_path, self.input_type, output_path, 'text', jpath="[*].[(annotations.organism || annotations.source), 'foo'] | [*].join('\t', @)")
        self.compare_files(Path.joinpath(self.output_path, 'gentype'), output_path)

    def test_creation(self):
        """
        Test handling JMESPath creating new SeqRecords
        """
        output_path = Path(self.workdir.name, 'faa')
        convert(self.input_path, self.input_type, output_path, 'fasta', jpath="""
[0].let({org: (annotations.organism || annotations.source)}, &(features[?type=='CDS' && qualifiers.translation].{id:
join('|', [
  (qualifiers.db_xref[?starts_with(@, 'GI')].['gi', split(':', @)[1]]),
  (qualifiers.protein_id[*].['ref', @]),
  (qualifiers.locus_tag[*].['locus', @]),
  join('', [':', [location][?strand==`-1`] && 'c' || '', to_string(sum([location.start, `1`])), '..', to_string(location.end)])
][][]),
seq: qualifiers.translation[0],
description: (org && join('', [qualifiers.product[0], ' [', org, ']']) || qualifiers.product[0])}))
        """)
        self.compare_files(Path.joinpath(self.output_path, 'faa'), output_path)

    def test_creation2(self):
        """
        Test handling JMESPath creating new SeqRecords
        """
        output_path = Path(self.workdir.name, 'ffn')
        convert(self.input_path, self.input_type, output_path, 'fasta', jpath="""
[0].let({desc: description, seq: seq}, &features[?type=='gene'].{id:
join('|', [
  (qualifiers.db_xref[?starts_with(@, 'GI')].['gi', split(':', @)[1]]),
  (qualifiers.protein_id[*].['ref', @]),
  (qualifiers.locus_tag[*].['locus', @]),
  join('', [':', [location][?strand==`-1`] && 'c' || '', to_string(sum([location.start, `1`])), '..', to_string(location.end)])
][][]),
seq: extract(seq, @),
description: desc})
        """)
        self.compare_files(Path.joinpath(self.output_path, 'ffn'), output_path)